@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 """
-A class for interactively annotating 2D matplotlib plots.
+A class for interactively annotating matplotlib plots.
 
 This class was inspired by code posted by HappyLeapSecond which was
 derived from code posted by Joe Kington. See:
@@ -9,18 +9,22 @@ http://stackoverflow.com/questions/13306519/\
 get-data-from-plot-with-matplotlib
 """
 
+from types import SimpleNamespace
 import numpy as np
 import matplotlib.pyplot as plt
+from matplotlib.collections import PathCollection
+from matplotlib.lines import Line2D
+from mpl_toolkits.mplot3d import proj3d
 
 
-def form1(x, y, n, ind, ax, line):
-    """Default annotation for plots with 1 line"""
-    return f"x: {x:.5g}\ny: {y:.5g}"
-
-
-def form2(x, y, n, ind, ax, line):
-    """Default annotation for plots with more than 1 line"""
-    return f"x: {x:.5g}\ny: {y:.5g}\n{line.get_label()}"
+def mk_label(point):
+    """Default annotation function"""
+    label = f"x: {point.x:.5g}\ny: {point.y:.5g}"
+    if point.z is not None:
+        label += f"\nz: {point.z:.5g}"
+    if point.nlines > 1:
+        label += f"\n{point.handle.get_label()}"
+    return label
 
 
 def _ensure_iterable(a):
@@ -48,23 +52,33 @@ class DataCursor(object):
             DC.off()
             DC.on()
 
-    form1 : function
-        Function to format the x, y data for the annotation. This is
-        called for axes that have only one line. Must accept the four
-        inputs listed and and return a string. The argument `n` is the
-        line number starting at 0 and `ind` is the index into the data
-        vectors. `form1` defaults to::
+    mk_label : function
+        Function that returns an annotation label for a data
+        point. Must accept a single SimpleNamespace argument that
+        contains information about the selected data point. `mk_label`
+        defaults to::
 
-          def form1(x, y, n, ind, ax, line):
-              return f"x: {x:.5g}\ny: {y:5g}"
+            def mk_label(point):
+                label = f"x: {point.x:.5g}\ny: {point.y:.5g}"
+                if point.z is not None:
+                    label += f"\nz: {point.z:.5g}"
+                if point.nlines > 1:
+                    label += f"\n{point.handle.get_label()}"
+                return label
 
-    form2 : function
-        Function to format the x, y data for the annotation. This is
-        called for axes that have more than one line. Same signature
-        as `form1`. Defaults to::
+        The SimpleNamespace ``point`` contains the following
+        attributes:
 
-          def form2(x, y, n, ind, ax, line):
-              return f"x: {x:.5g}\ny: {y:.5g}\n{line.get_label()}"
+        ====== ========================================================
+         attr  Description
+        ====== ========================================================
+        ax     Axes handle
+        handle Line2D or PathCollection handle for line
+        index  Index into the data vectors for data point
+        lines  Total number of lines on axes
+        n      Line number starting at 0
+        x,y,z  x, y, z coordinates of data point; z will be None for 2D
+        ====== ========================================================
 
     offsets : tuple
         Two element tuple containing x and y offsets in points for the
@@ -84,23 +98,19 @@ class DataCursor(object):
     permdot : dict; optional
         Similar to `followdot` except this is a "permanent" dot;
         this gets placed after left clicking.
-    xypoints : list
-        Contains [x, y] data pairs for each (non-deleted) left-click.
-    inds : list
-        Contains index of [x, y] data pairs for each (non-deleted)
-        left-click. Same length as `xypoints`.
-    lines : list
-        Contains line object handles for the selected lines. Same
-        length as `xypoints`.
-    linenums : list
-        Contains line number for each [x, y] data pair, starting at 0.
-        Same length as `xypoints`.
+    points : list
+        Contains list of SimpleNamespace objects. Each "point" is as
+        described above under the `mk_lable` attribute.
     pts : list
-        Contains the :func:`matplotlib.pyplot.scatter` object handles
-        for the selected points. Same length as `xypoints`.
+        Contains the :func:`matplotlib.pyplot.scatter`
+        (PathCollection) object handles for the selected and annotated
+        points. Same length as `points`. Note that
+        :class:`DataCursor` ignores these added annotation points when
+        moused over; they are identified by the added attribute
+        "_pyyeti_dc_point".
     notes : list
         Contains the annotation object handles for the selected points.
-        Same length as `xypoints`.
+        Same length as `points`.
 
     Notes
     -----
@@ -129,32 +139,28 @@ class DataCursor(object):
     Action       Description
     ===========  =====================================================
     left-click   Data point will be stored in the member list
-                 `xypoints`. The other members (`inds`, `lines`,
-                 `linenums`, `pts` and `notes`) are updated as well.
-    right-click  Last point is deleted from `xypoints` and from the
+                 `points`. The `pts` and `notes` attributes are
+                 updated as well.
+    right-click  Last point is deleted from `points` and from the
                  other members.
     typing 't'   Turns off DataCursor. To turn on, use ``DC.on``. Note
-                 that turning on will reset `xypoints` and the other
+                 that turning on will reset `points` and the other
                  members.
     typing 'D'   Deletes last point AND removes the line from the plot
                  via ``line_handle.remove()``. Any older annotations
                  are not deleted.
     ===========  =====================================================
 
-    To get x-y data points from plots from within a script, use
+    To get data points from plots from within a script, use
     :func:`DataCursor.getdata`. Enter the number of points or press
     't' to end blocking so the script will continue (see
     :func:`DataCursor.getdata`).
 
     Once the DataCursor is turned off, the annotations become
-    draggable. Two notes on this:
-
-        1. Due to a bug in matplotlib v1.4.3, dragging annotations
-           connected to points with negative x or y coordinates
-           doesn't work correctly. Newer versions work properly.
-        2. Somehow, annotations sometimes become linked (moving one
-           will move another). When that happens, try dragging
-           a different annotation; this sometimes breaks the link.
+    draggable. Note that, at least for some versions of Matplotlib,
+    annotations sometimes become linked (moving one will move
+    another). When that happens, try dragging a different annotation;
+    this sometimes breaks the link.
 
     Interactively, the member functions :func:`DataCursor.on` and
     :func:`DataCursor.off` are used to turn the DataCursor on and
@@ -180,21 +186,23 @@ class DataCursor(object):
         # use DC.pause if you want to drag boxes around before
         # continuing:
         DC.pause()
-        print('points are:', DC.xypoints)
+        print('x, y values of selected points are:')
+        print(np.array([[p.x, p.y] for p in DC.points]))
 
     Settings can be changed after instantiation. Here is an example of
-    defining a new format for the annotation (for both `form1` and
-    `form2`). Only the line label is included in the annotation. The
-    example also changes the permanent dot to a gray pentagon::
+    defining a new format for the annotation. Only the line label is
+    included in the annotation. The example also changes the permanent
+    dot to a gray pentagon::
 
         import matplotlib.pyplot as plt
         import numpy as np
         from pyyeti.datacursor import DC
 
-        def formnew(x, y, n, ind, ax, line):
-            return f'{line.get_label()}'
+        def new_label(point):
+            return (f'{point.handle.get_label()}\n'
+                    f'({point.x},{point.y:.2f})')
 
-        DC.form1 = DC.form2 = formnew
+        DC.mk_label = new_label
         DC.permdot = dict(s=130, color='black', alpha=0.4,
                           marker='p')
 
@@ -213,16 +221,20 @@ class DataCursor(object):
         import numpy as np
         from pyyeti.datacursor import DC
 
-        def addpt(x, y, n, ind, ax, line):
-            print(f'You selected ({x}, {y})')
+        def addpt(point):
+            print(f'You selected ({point.x}, {point.y}, {point.z})')
 
-        def delpt(x, y, n, ind, ax, line):
-            print(f'You deleted ({x}, {y})')
+        def delpt(point):
+            print(f'You deleted ({point.x}, {point.y}, {point.z})')
 
         DC.addpt_func(addpt)
         DC.delpt_func(delpt)
 
-        plt.plot(np.random.randn(50))
+        fig = plt.figure()
+        ax = fig.add_subplot(projection="3d")
+        coords = np.random.randn(6, 3)
+        dots = ax.scatter(*coords.T)
+        ax.plot(*(coords.T + 0.1), "v")
         DC.on()
     """
 
@@ -231,8 +243,7 @@ class DataCursor(object):
         ax=None,
         figs=None,
         hover=True,
-        form1=form1,
-        form2=form2,
+        mk_label=mk_label,
         offsets=(-20, 20),
         bbox=dict(boxstyle="round,pad=0.5", fc="gray", alpha=0.5),
         arrowprops=dict(arrowstyle="->", connectionstyle="arc3,rad=0"),
@@ -274,8 +285,7 @@ class DataCursor(object):
         self._ax_input = ax
         self._figs_input = figs
         self.hover = hover
-        self.form1 = form1
-        self.form2 = form2
+        self.mk_label = mk_label
         self.offsets = offsets
         self.bbox = bbox
         self.arrowprops = arrowprops
@@ -311,8 +321,9 @@ class DataCursor(object):
                 return
         maxlines = 0
         for a in self._ax:
-            if len(a.lines) > maxlines:
-                maxlines = len(a.lines)
+            handles = self._get_data_handles(a)
+            if len(handles) > maxlines:
+                maxlines = len(handles)
         if maxlines == 0 and errout:
             raise RuntimeError("no lines; plot something first")
 
@@ -341,7 +352,7 @@ class DataCursor(object):
             call-back is turned on (used for pausing; see
             :func:`pause`).
         reset : bool; optional
-            If True, the `xypoints` and other data members are reset
+            If True, the `points` and other data members are reset
             to empty lists. Otherwise, if `reset` is False, your new
             data will be appended on to your previous data.
 
@@ -355,10 +366,7 @@ class DataCursor(object):
             self._figs_input = figs
         self._init_all()
         if reset:
-            self.xypoints = []
-            self.inds = []
-            self.lines = []
-            self.linenums = []
+            self.points = []
             self.pts = []
             self.notes = []
         self._kid = {}  # key press event
@@ -428,23 +436,10 @@ class DataCursor(object):
         """
         Function to call on a left-click.
 
-        Call signature is: ``func(x, y, n, ind, ax, line)``
+        Call signature is: ``func(point)``
 
-        The parameters for the function are described below.
-
-        Parameters
-        ----------
-        x, y : scalars
-            The selected ``(x, y)`` data point
-        n : integer
-            The line number
-        ind : integer
-            The index of the data point within the plotted data
-            vectors
-        ax : axes object
-            Axes handle
-        line : line object
-            Line handle
+        The input ``point`` is a SimpleNamespace as described for the
+        `mk_label` attribute. See :class:`DataCursor`.
         """
         self.addptfunc = func
 
@@ -452,50 +447,42 @@ class DataCursor(object):
         """
         Function to call on a right-click.
 
-        Call signature is: ``func(x, y, n, ind, ax, line)``
+        Call signature is: ``func(point)``
 
-        The parameters for the function are described below.
-
-        Parameters
-        ----------
-        x, y : scalars
-            The selected ``(x, y)`` data point
-        n : integer
-            The line number
-        ind : integer
-            The index of the data point within the plotted data
-            vectors
-        ax : axes object
-            Axes handle
-        line : line object
-            Line handle or None; None if the line was deleted
+        The input ``point`` is a SimpleNamespace as described for the
+        `mk_label` attribute. See :class:`DataCursor`.
         """
         self.delptfunc = func
 
-    def _add_point(self, x, y, n, ind, ax, line):
-        self.xypoints.append([x, y])
-        self.inds.append(ind)
-        self.lines.append(line)
-        self.linenums.append(n)
+    @staticmethod
+    def _add_annotation_point(ax, x, y, dct, vis):
+        h = ax.scatter(x, y, **dct, visible=vis)
+        h._pyyeti_dc_point = True
+        return h
+
+    @staticmethod
+    def _is3d(ax):
+        return hasattr(ax, "get_proj")
+
+    def _add_point(self, x, y, point):
+        self.points.append(point)
         # if 3d, set visible to False (i don't know how to get proper
         # coordinates for the dot):
-        vis = False if hasattr(ax, "get_proj") else True
-        self.pts.append(ax.scatter(x, y, **self.permdot, visible=vis))
-        self.notes.append(self._annotation[ax])
+        vis = False if self._is3d(point.ax) else True
+        self.pts.append(self._add_annotation_point(point.ax, x, y, self.permdot, vis))
+        self.notes.append(self._annotation[point.ax])
         # make a new annotation box so current one is static
-        self._annotation[ax] = self._new_annotation(ax, (x, y))
-        if self._in_loop and len(self.xypoints) == self._max_points:
+        self._annotation[point.ax] = self._new_annotation(point.ax, (x, y))
+        if self._in_loop and len(self.points) == self._max_points:
             self.off()
         if self.addptfunc:
-            self.addptfunc(x, y, n, ind, ax, line)
+            self.addptfunc(point)
 
     def _del_point(self, ax=None, delete_line=False):
         """Deletes last saved point, if any."""
-        if len(self.xypoints) == 0:
+        if len(self.points) == 0:
             return
-        x, y = self.xypoints.pop()
-        ind = self.inds.pop()
-        h = self.lines.pop()
+        point = self.points.pop()
         if delete_line:
             # line may have been deleted already, so catch exception:
             try:
@@ -504,13 +491,12 @@ class DataCursor(object):
                 pass
             else:
                 h = None
-        n = self.linenums.pop()
         pt = self.pts.pop()
         pt.remove()
         na = self.notes.pop()
         na.remove()
         if self.delptfunc:
-            self.delptfunc(x, y, n, ind, ax, h)
+            self.delptfunc(point)
 
     def _get_ax(self, event):
         """
@@ -564,27 +550,22 @@ class DataCursor(object):
         ax = self._get_ax(event)
         if not ax:
             return
-        x, y = event.xdata, event.ydata
-        x, y, n, ind, line = self._snap(ax, x, y)
+        x, y, point = self._snap(ax, event.xdata, event.ydata)
         if x is None:
             return
         annotation = self._annotation[ax]
         dot = self._dot[ax]
         annotation.xy = x, y
-        if len(ax.lines) > 1:
-            annotation.set_text(self.form2(x, y, n, ind, ax, line))
-        else:
-            annotation.set_text(self.form1(x, y, n, ind, ax, line))
+        annotation.set_text(self.mk_label(point))
         dot.set_offsets((x, y))
         if event.name == "button_press_event":
             if event.button == 1:
-                # dot.set_visible(True)
                 annotation.set_visible(True)
-                self._add_point(x, y, n, ind, ax, line)
-            elif event.button == 3 and len(self.xypoints) > 0:
+                self._add_point(x, y, point)
+            elif event.button == 3 and len(self.points) > 0:
                 self._del_point(ax=ax)
         elif self.hover:
-            if not hasattr(ax, "get_proj"):
+            if not self._is3d(ax):
                 dot.set_visible(True)
             annotation.set_visible(True)
         event.canvas.draw()
@@ -602,38 +583,86 @@ class DataCursor(object):
             visible=False,
         )
 
+    @staticmethod
+    def _get_data_handles(ax):
+        return [
+            child
+            for child in ax.get_children()
+            if isinstance(child, (PathCollection, Line2D))
+            and not hasattr(child, "_pyyeti_dc_point")
+        ]
+
+    @staticmethod
+    def _get_xy_data(ax, h):
+        if isinstance(h, Line2D):
+            return h.get_xdata().astype(float), h.get_ydata().astype(float)
+
+        if hasattr(h, "_offsets3d"):
+            x, y, _ = proj3d.proj_transform(*np.array(h._offsets3d), ax.get_proj())
+            return x, y
+            # return (*h.get_offsets().T,) # returns in unknown order
+
+        return (*h.get_offsets().data.T,)
+
+    @staticmethod
+    def _get_xyz(ax, h, ind):
+        """
+        Get original 3D location of a point, instead of the mapped-to-2D
+        x, y location
+        """
+        if isinstance(h, Line2D):
+            return np.array(h.get_data_3d())[:, ind]
+        return np.array(h._offsets3d)[:, ind]
+
     def _setup_annotations(self):
         """Create the annotation boxes. The string value and the
         position will be set for each event."""
         self._annotation = {}
         self._dot = {}
         for ax in self._ax:
-            if len(ax.lines) > 0:
-                ln = ax.lines[0]
-                xy = ln.get_xdata()[0], ln.get_ydata()[0]
+            handles = self._get_data_handles(ax)
+            if len(handles) > 0:
+                x, y = self._get_xy_data(ax, handles[0])
+                xy = x[0], y[0]
             else:
                 xy = 0, 0
             self._annotation[ax] = self._new_annotation(ax, xy)
             xl, yl = ax.get_xlim(), ax.get_ylim()
-            self._dot[ax] = ax.scatter(xy[0], xy[1], **self.followdot, visible=False)
+            self._dot[ax] = self._add_annotation_point(ax, *xy, self.followdot, False)
             ax.set_xlim(xl)
             ax.set_ylim(yl)
 
     def _snap(self, ax, x, y):
-        """Return the value in self._points closest to x, y."""
+        """Return the plotted value closest to x, y."""
         dmin = np.inf
-        xyn = None, None, None, None, None
-        for n, ln in enumerate(ax.lines):
-            xdata = ln.get_xdata().astype(float)
-            ydata = ln.get_ydata().astype(float)
+        lines = self._get_data_handles(ax)
+        nlines = len(self._get_data_handles(ax))
+        if nlines == 0:
+            return None, None, None
+        for n, h in enumerate(lines):
+            xdata, ydata = self._get_xy_data(ax, h)
             dx = (xdata - x) / np.diff(ax.get_xlim())[0]
             dy = (ydata - y) / np.diff(ax.get_ylim())[0]
-            d = np.sqrt(dx ** 2.0 + dy ** 2.0)
+            d = dx ** 2.0 + dy ** 2.0
             ind = np.argmin(d)
             if d[ind] < dmin:
                 dmin = d[ind]
-                xyn = xdata[ind], ydata[ind], n, ind, ln
-        return xyn
+                best = xdata[ind], ydata[ind], n, ind, h
+
+        # x, y, n, ind, handle = self._snap(ax, x, y)
+        if best is None:
+            return None, None, None
+        x, y, n, ind, handle = best
+        if self._is3d(ax):
+            # get original data coordinates
+            xo, yo, zo = self._get_xyz(ax, handle, ind)
+        else:
+            xo, yo = x, y
+            zo = None
+        point = SimpleNamespace(
+            x=xo, y=yo, z=zo, ax=ax, n=n, index=ind, handle=handle, nlines=nlines
+        )
+        return x, y, point
 
     def _fake_getdata(self):
         """
